#synthDrivers/espeak.py
# -*- encoding: cp932 -*-
#A part of NonVisual Desktop Access (NVDA)
#Copyright (C) 2006-2009 NVDA Contributors <http://www.nvda-project.org/>
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.

import os
from collections import OrderedDict
import _espeak
import Queue
import threading
import languageHandler
from synthDriverHandler import SynthDriver,VoiceInfo,BooleanSynthSetting, NumericSynthSetting, SynthSetting
import speech
from logHandler import log
import config

class SynthDriver(SynthDriver):
	name = "ptenglish"
	description = "ProTALKER english"

	supportedSettings=(
		SynthDriver.VoiceSetting(),
		SynthDriver.VariantSetting(),
		SynthDriver.RateSetting(),
		# Translators: This is the name of the rate boost voice toggle
		# which further increases the speaking rate when enabled.
		BooleanSynthSetting("rateBoost",_("Rate boos&t")),
		SynthDriver.PitchSetting(),
		SynthDriver.InflectionSetting(),
		SynthDriver.VolumeSetting(),
		NumericSynthSetting("ptRate", _("ProTALKER rate"),minStep=10, normalStep=10),
		NumericSynthSetting("ptRateBoost", _("ProTALKER rate boost"), normalStep=10),
	SynthSetting("ptVoice", _("ProTALKER Voice")),
	)

	availablePtvoices = OrderedDict()
	availablePtvoices[u"1"] = VoiceInfo(u"1", u"ǂݏGN", u"ja")
	availablePtvoices[u"2"] = VoiceInfo(u"2", u"ǂݎq", u"ja")

	@classmethod
	def check(cls):
		return True

	def __init__(self):
		_espeak.initialize()
		log.info("Using eSpeak version %s" % _espeak.info())
		lang=languageHandler.getLanguage()
		_espeak.setVoiceByLanguage(lang)
		self._language=lang
		self._variantDict=_espeak.getVariantDict()
		self.variant="max"
		self.rate=30
		self.pitch=40
		self.inflection=75
		self.ptRate = 100
		self.ptRateBoost = 0
#For old configs, this will get set to 50. Fix that
		if 'ptenglish' in config.conf['speech'] and 'ptRateBoost' not in config.conf['speech']['ptenglish']:
			config.conf['speech']['ptenglish']['ptRateBoost'] = 0
		self.ptVoice = 1

	def _get_language(self):
		return self._language

	def speak(self,speechSequence):
		defaultLanguage=self._language
		textList=[]
		langChanged=False
		newlang = ''
		speechSequence = modseq(speechSequence, defaultLanguage)
		for item in speechSequence:
			if isinstance(item,basestring):
				s=unicode(item)
				if newlang == 'ja':
					if textList and textList[0] == '<voice xml:lang="ja">':
						del textList[0]
					else:
						_espeak.speak(u"".join(textList))
					textList = [u"</voice>"]
					textList.append("<voice xml:lang=\"%s\">"%(defaultLanguage).replace('_','-'))
					langChanged = True
					for item2 in item.replace(u'\u3000', ' ').split(' '):
						_espeak.ptspeak(item2)
					continue
				# Replace \01, as this is used for embedded commands.
				#Also replace < and > as espeak handles xml
				s.translate({ord(u'\01'):None,ord(u'<'):u'&lt;',ord(u'>'):u'&gt;'})
				textList.append(s)
			elif isinstance(item,speech.IndexCommand):
				textList.append("<mark name=\"%d\" />"%item.index)
			elif isinstance(item,speech.CharacterModeCommand):
				textList.append("<say-as type=\"spell-out\">" if item.state else "</say-as>")
			elif isinstance(item,speech.LangChangeCommand):
				newlang = item.lang
				if langChanged:
					textList.append("</voice>")
				textList.append("<voice xml:lang=\"%s\">"%(item.lang if item.lang else defaultLanguage).replace('_','-'))
				langChanged=True
			elif isinstance(item,speech.SpeechCommand):
				log.debugWarning("Unsupported speech command: %s"%item)
			else:
				log.error("Unknown speech: %s"%item)
		if langChanged:
			textList.append("</voice>")
		text=u"".join(textList)
		_espeak.speak(text)

	def cancel(self):
		_espeak.stop()

	def pause(self,switch):
		_espeak.pause(switch)

	_rateBoost = False
	RATE_BOOST_MULTIPLIER = 3

	def _get_rateBoost(self):
		return self._rateBoost

	def _set_rateBoost(self, enable):
		if enable == self._rateBoost:
			return
		rate = self.rate
		self._rateBoost = enable
		self.rate = rate

	def _get_rate(self):
		val=_espeak.getParameter(_espeak.espeakRATE,1)
		if self._rateBoost:
			val=int(val/self.RATE_BOOST_MULTIPLIER)
		return self._paramToPercent(val,_espeak.minRate,_espeak.maxRate)

	def _set_rate(self,rate):
		val=self._percentToParam(rate, _espeak.minRate, _espeak.maxRate)
		if self._rateBoost:
			val=int(val*self.RATE_BOOST_MULTIPLIER)
		_espeak.setParameter(_espeak.espeakRATE,val,0)

	def _get_pitch(self):
		val=_espeak.getParameter(_espeak.espeakPITCH,1)
		return self._paramToPercent(val,_espeak.minPitch,_espeak.maxPitch)

	def _set_pitch(self,pitch):
		val=self._percentToParam(pitch, _espeak.minPitch, _espeak.maxPitch)
		_espeak.setParameter(_espeak.espeakPITCH,val,0)

	def _get_inflection(self):
		val=_espeak.getParameter(_espeak.espeakRANGE,1)
		return self._paramToPercent(val,_espeak.minPitch,_espeak.maxPitch)

	def _set_inflection(self,val):
		val=self._percentToParam(val, _espeak.minPitch, _espeak.maxPitch)
		_espeak.setParameter(_espeak.espeakRANGE,val,0)

	def _get_volume(self):
		return _espeak.getParameter(_espeak.espeakVOLUME,1)

	def _set_volume(self,volume):
		_espeak.setParameter(_espeak.espeakVOLUME,volume,0)

	def _getAvailableVoices(self):
		voices=OrderedDict()
		for v in _espeak.getVoiceList():
			l=v.languages[1:]
			identifier=os.path.basename(v.identifier)
			voices[identifier]=VoiceInfo(identifier,v.name,l)
		return voices

	def _get_voice(self):
		curVoice=getattr(self,'_voice',None)
		if curVoice: return curVoice
		curVoice = _espeak.getCurrentVoice()
		if not curVoice:
			return ""
		return curVoice.identifier.split('+')[0]

	def _set_voice(self, identifier):
		if not identifier:
			return
		if "\\" in identifier:
			identifier=os.path.basename(identifier)
		self._voice=identifier
		try:
			_espeak.setVoiceAndVariant(voice=identifier,variant=self._variant)
		except:
			self._voice=None
			raise
		self._language=super(SynthDriver,self).language

	def _get_lastIndex(self):
		return _espeak.lastIndex

	def terminate(self):
		_espeak.terminate()

	def _get_variant(self):
		return self._variant

	def _set_variant(self,val):
		self._variant = val if val in self._variantDict else "max"
		_espeak.setVoiceAndVariant(variant=self._variant)

	def _getAvailableVariants(self):
		return OrderedDict((ID,VoiceInfo(ID, name)) for ID, name in self._variantDict.iteritems())

	def _get_ptRate(self):
		return self._paramToPercent(_espeak.pt.rate, 0, 9)

	def _set_ptRate(self, value):
		_espeak.pt.rate = self._percentToParam(value, 0, 9)

	def _get_ptRateBoost(self):
		return self._paramToPercent(_espeak.pt.boost, 100, 200)

	def _set_ptRateBoost(self, value):
		_espeak.pt.boost = self._percentToParam(value, 100, 200)

	#For NVDAJP's extended NVDAHelper
	def isSpeaking(self):
		global isSpeaking
		return _espeak.isSpeaking

	def _get_ptVoice(self):
		return str(_espeak.pt.params.font)

	def _set_ptVoice(self, val):
		_espeak.pt.params.font = int(val)

def stringsplit(s, lang):
 if s.strip() == '': return []
 kinds = (speech.LangChangeCommand(lang), speech.LangChangeCommand('ja'))
 def kind_of(u):
  """Returns kind of character, which is represented by a unicode codepoint. 0 for english, 1 for japanese."""
  if (u >= 0x3000 and u <= 0x30ff) or (u >= 0x4e00 and u <= 0x9fbf) or (u >= 0xff00 and u <= 0xffef) or (u >= 0x2160 and u <= 0x2169): return 1
  return 0
 lst = []
 lastkind = 0
 start = 0
 #If lastkind is not set correctly here, the first entry will be blank.
 lastkind = kind_of(ord(s[0]))

 for pos, c in enumerate(s):
  u = ord(c)
  if u == 32: continue #spaces don't change anything
  kind = kind_of(u)
  if kind != lastkind: #switched languages
   lst.extend([kinds[lastkind], s[start:pos]])
   lastkind = kind
   start = pos

 #The final piece of text that wasn't inserted yet
 lst.extend([kinds[kind], s[start:pos+1]])
 return lst
def modseq(seq, lang):
	newseq = []
	for item in seq:
		if isinstance(item, speech.LangChangeCommand):
			continue
		if isinstance(item, basestring):
			item = unicode(item).translate(jpn_translate)
			newseq.extend(stringsplit(unicode(item), lang))
		else:
			newseq.append(item)
	return newseq
jpn_translate = {
0x2170: 0x2160,
0x2171: 0x2161,
0x2172: 0x2162,
0x2173: 0x2163,
0x2174: 0x2164,
0x2175: 0x2165,
0x2176: 0x2166,
0x2177: 0x2167,
0x2178: 0x2168,
0x2179: 0x2169,
}